/*-----------------------------------------------------------------------------
 * $RCSfile: Serializer.c,v $
 *
 * See Copyright for the status of this software.
 *
 * The OpenSOAP Project
 * http://opensoap.jp/
 *-----------------------------------------------------------------------------
 */

#include "Serializer.h"

/*
=begin
= OpenSOAP Serializer class
=end
 */

/*
=begin
--- function#OPENSOAP_SERIALIZER_NAME(bool)(*from_value, to)
    OpenSOAP Serializer of Boolean Type

    :Parameters
      :[in]  void * ((|from_value|))
        Serializer Object Assign Pointer
      :[out] OpenSOAPStringPtr ((|to|))
	    Result Value Pointer
    :Return Value
      :int
	    Return Code Is ErrorCode
    :Note
      Boolean^,INTƂĈ     0,1,True,False
=end
 */
static
const wchar_t *
SOAPTrues[] = {
    L"true", L"1", NULL
};

static
const wchar_t *
SOAPFalses[] = {
    L"false", L"0", NULL
};

OPENSOAP_SERIALIZER_DEFINITION(boolean) {
    int ret = OPENSOAP_PARAMETER_BADVALUE;
 
    if (from_value && to) {
        ret = OpenSOAPStringSetStringWC(to,
                                        (*(int *)from_value)
                                        ? SOAPTrues[0]
                                        : SOAPFalses[0]);
    }

    return ret;
}

/*
=begin
--- function#OPENSOAP_DESERIALIZER_NAME(bool)(from, to_value)
    OpenSOAP Deserializer of Boolean Type

    :Parameters
      :[in]  OpenSOAPStringPtr ((|from|))
        Deserializer Object Assign Pointer
      :[out] void * ((|to_value|))
	    Result Value Pointer
    :Return Value
      :int
	    Return Code Is ErrorCode
    :Note
      Boolean^,INTƂĈ     0,1,True,False
=end
 */
OPENSOAP_DESERIALIZER_DEFINITION(boolean) {
    int ret = OPENSOAP_PARAMETER_BADVALUE;

    if (from && to_value) {
        int cmp_rslt = 0;
        const wchar_t **cmp_strs = SOAPTrues;

        for (; *cmp_strs; ++cmp_strs) {
            ret = OpenSOAPStringCompareWC(from,
                                          *cmp_strs,
                                          &cmp_rslt);
            if (OPENSOAP_FAILED(ret) || cmp_rslt == 0) {
                break;
            }
        }
        if (OPENSOAP_SUCCEEDED(ret)) {
            if (*cmp_strs) {
                *(int *)to_value = 1;
            }
            else {
                for (cmp_strs = SOAPFalses; *cmp_strs; ++cmp_strs) {
                    ret = OpenSOAPStringCompareWC(from,
                                                  *cmp_strs,
                                                  &cmp_rslt);
                    if (OPENSOAP_FAILED(ret) || cmp_rslt == 0) {
                        break;
                    }
                }
                if (OPENSOAP_SUCCEEDED(ret) && *cmp_strs) {
                    *(int *)to_value = 0;
                }
            }
        }
    }

    return ret;
}

/* Base64 encoding/decoding */

#include <string.h>
#include <ctype.h>

static
const unsigned char
base64Code[]
= "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/=";

static
const unsigned char
base64PadChar = '=';

static
unsigned char
getBase64Code(unsigned char c) {
	unsigned char ret = (unsigned char)(-1);
	const unsigned char *search
		= strchr(base64Code, c);
	if (search) {
		ret = (search - base64Code);
	}

	return ret;
}

/*
=begin
--- function#OPENSOAP_SERIALIZER_NAME(base64Binary)(from_value, to)
    OpenSOAP Serializer of Bibary(Base64Binary) Type

    :Parameters
      :[in]  void * ((|from_value|))
        Serializer Object Assign Pointer
      :[out] OpenSOAPStringPtr ((|to|))
	    Result Value Pointer
    :Return Value
      :int
	    Return Code Is ErrorCode
    :Note
      Binary^(Base64Binary)ByteArray
=end
 */
OPENSOAP_SERIALIZER_DEFINITION(base64Binary) {
    int ret = OPENSOAP_PARAMETER_BADVALUE;

	if (from_value && to) {
		OpenSOAPByteArrayPtr from_bary = *(OpenSOAPByteArrayPtr *)from_value;
		const unsigned char *from_value_beg = NULL;
		size_t from_sz = 0;
		ret = OpenSOAPByteArrayGetBeginSizeConst(from_bary,
												 &from_value_beg,
												 &from_sz);
		if (OPENSOAP_SUCCEEDED(ret)) {
			OpenSOAPByteArrayPtr to_str = NULL;
			ret = OpenSOAPByteArrayCreate(&to_str);
			if (OPENSOAP_SUCCEEDED(ret)) {
#define	B64UNIT_SIZE 3
				size_t	i = 0;
				size_t	padding_sz
					= (from_sz + B64UNIT_SIZE - 1)
					/ B64UNIT_SIZE * B64UNIT_SIZE;
				while (OPENSOAP_SUCCEEDED(ret) && i != padding_sz) {
#define	B64APPEND_SIZE 4
					unsigned char append_buf[B64APPEND_SIZE];
					size_t next_i = i + B64UNIT_SIZE;
					size_t append_from_sz
						= (next_i == padding_sz)
						? (from_sz - i) : B64UNIT_SIZE;
					const unsigned char *from_i
						= from_value_beg + i;
					unsigned char *a_i = append_buf;
					
					*a_i++ = base64Code[(from_i[0] & 0xfc) >> 2];
					*a_i++ = base64Code[((from_i[0] & 0x03) << 4)
										 | ((append_from_sz > 1)
											? ((from_i[1] & 0xf0) >> 4)
											: '\0')];
					*a_i++ = (append_from_sz > 1)
						? base64Code[((from_i[1] & 0x0f) << 2)
									  | ((append_from_sz > 2)
										 ? ((from_i[2] & 0xc0) >> 6)
										  : '\0')]
						: base64PadChar;
					*a_i++ = (append_from_sz > 2)
						? base64Code[(from_i[2] & 0x3f)] : base64PadChar;

					ret = OpenSOAPByteArrayAppend(to_str,
												  append_buf,
												  B64APPEND_SIZE);
					i = next_i;
#undef	B64APPEND_SIZE
				}

				if (OPENSOAP_SUCCEEDED(ret)) {
					ret = OpenSOAPStringSetCharEncodingString(to,
															  "US-ASCII",
															  to_str);
				}
				OpenSOAPByteArrayRelease(to_str);
#undef	B64UNIT_SIZE
			}
		}
	}

    return ret;
}

static
int
OpenSOAPDeserializerBase64(/* [in] */ OpenSOAPByteArrayPtr from,
						   /* [out] */ OpenSOAPByteArrayPtr to) {
    int ret = OPENSOAP_PARAMETER_BADVALUE;

    if (from && to) {
		ret = OpenSOAPByteArrayClear(to);
		if (OPENSOAP_SUCCEEDED(ret)) {
			const unsigned char *from_value = NULL;
			size_t	from_sz = 0;
			ret = OpenSOAPByteArrayGetBeginSizeConst(from,
													 &from_value,
													 &from_sz);
			if (OPENSOAP_SUCCEEDED(ret)) {
#define	B64DECUNIT_SIZE 4
				unsigned char decunit[B64DECUNIT_SIZE];
				unsigned char *d_i = decunit;
				size_t i = 0;
				while (i != from_sz && OPENSOAP_SUCCEEDED(ret)) {
					unsigned char c = from_value[i++];
					if (!isspace(c)) {
						unsigned char d = getBase64Code(c);
						if (d != (unsigned char)(-1)) {
							*d_i++ = d;
						}
						else {
							ret = OPENSOAP_PARAMETER_BADVALUE;
							break;
						}
					}
					if (d_i == &decunit[B64DECUNIT_SIZE]) {
#define	B64UNIT_SIZE 3
						unsigned char append_data[B64UNIT_SIZE];
						size_t append_size = 0;
					
						append_data[append_size++]
							= ((decunit[0] & 0x3f) << 2)
							| ((decunit[1] & 0x30) >> 4);
						if (decunit[2] < 0x40) {
							append_data[append_size++]
								= ((decunit[1] & 0x0f) << 4)
								| ((decunit[2] & 0x3c) >> 2);
							if (decunit[3] < 0x40) {
								append_data[append_size++]
									= ((decunit[2] & 0x03) << 6)
									| (decunit[3] & 0x3f);
							}
						}
					
						ret = OpenSOAPByteArrayAppend(to,
													  append_data,
													  append_size);
						d_i = decunit;
#undef	B64UNIT_SIZE
					}
				}
#undef	B64DECUNIT_SIZE
			}
		}
	}

	return ret;
}

/*
=begin
--- function#OPENSOAP_DESERIALIZER_NAME(base64Binary)(from, to_value)
    OpenSOAP Deserializer of Bibary(Base64Binary) Type

    :Parameters
      :[in]  void * ((|from_value|))
        Deserializer Object Assign Pointer
      :[out] OpenSOAPStringPtr ((|to|))
	    Result Value Pointer
    :Return Value
      :int
	    Return Code Is ErrorCode
    :Note
      Binary^(Base64Binary)ByteArray
=end
 */
OPENSOAP_DESERIALIZER_DEFINITION(base64Binary) {
    int ret = OPENSOAP_PARAMETER_BADVALUE;

    if (from && to_value) {
		OpenSOAPByteArrayPtr	from_bary = NULL;
		ret = OpenSOAPByteArrayCreate(&from_bary);
		if (OPENSOAP_SUCCEEDED(ret)) {
			ret = OpenSOAPStringGetCharEncodingString(from,
													  "US-ASCII",
													  from_bary);
			if (OPENSOAP_SUCCEEDED(ret)) {
				ret = OpenSOAPDeserializerBase64(from_bary,
												 *(OpenSOAPByteArrayPtr *)
												 to_value);
			}

			OpenSOAPByteArrayRelease(from_bary);
		}
    }

    return ret;
}

/*
=begin
--- function#OPENSOAP_SERIALIZER_NAME(double)(from_value, to)
    OpenSOAP Serializer of Double Type

    :Parameters
      :[in]  void * ((|from_value|))
        Serializer Object Assign Pointer
      :[out] OpenSOAPStringPtr ((|to|))
	    Result Value Pointer
    :Return Value
      :int
	    Return Code Is ErrorCode
    :Note
      64bit flowting
=end
 */
OPENSOAP_SERIALIZER_DEFINITION(double) {
    int ret = OPENSOAP_PARAMETER_BADVALUE;

    if (from_value && to) {
        double val = *(double *)from_value;
        ret = OpenSOAPStringFormatWC(to,
                                     L"%f",
                                     val);
    }
    return ret;
}

/*
=begin
--- function#OPENSOAP_DESERIALIZER_NAME(double)(from, to_value)
    OpenSOAP Deserializer of Double Type

    :Parameters
      :[in]  OpenSOAPStringPtr ((|from|))
        Deserializer Object Assign Pointer
      :[out] void * ((|to_value|))
	    Result Value Pointer
    :Return Value
      :int
	    Return Code Is ErrorCode
    :Note
      64bit flowting
=end
 */
OPENSOAP_DESERIALIZER_DEFINITION(double) {
    int ret = OPENSOAP_PARAMETER_BADVALUE;

    if (from && to_value) {
		char *buf = NULL;
		ret = OpenSOAPStringGetStringMBWithAllocator(from,
													 NULL, /* use malloc */
													 NULL, /* not use length */
													 &buf);
        if (OPENSOAP_SUCCEEDED(ret)) {
			*(double *)to_value = atof(buf);
			free(buf);
        }
    }
    return ret;
}

/*
=begin
--- function#OPENSOAP_SERIALIZER_NAME(int)(from_value, to)
    OpenSOAP Serializer of Integer Type

    :Parameters
      :[in]  void * ((|from_value|))
        Serializer Object Assign Pointer
      :[out] OpenSOAPStringPtr ((|to|))
	    Result Value Pointer
    :Return Value
      :int
	    Return Code Is ErrorCode
    :Note
      32bit
=end
 */
OPENSOAP_SERIALIZER_DEFINITION(int) {
    int ret = OPENSOAP_PARAMETER_BADVALUE;

    if (from_value && to) {
        long val = *(long *)from_value;
        ret = OpenSOAPStringFormatWC(to,
                                     L"%ld",
                                     val);
    }
    
    return ret;
}

/*
=begin
--- function#OPENSOAP_DESERIALIZER_NAME(int)(from, to_value)
    OpenSOAP Deserializer of Integer Type

    :Parameters
      :[in]  OpenSOAPStringPtr ((|from|))
        Serializer Object Assign Pointer
      :[out] void * ((|to_value|))
	    Result Value Pointer
    :Return Value
      :int
	    Return Code Is ErrorCode
    :Note
      32bit
=end
 */
OPENSOAP_DESERIALIZER_DEFINITION(int) {
    int ret = OPENSOAP_PARAMETER_BADVALUE;

    if (from && to_value) {
		char *buf = NULL;
		ret = OpenSOAPStringGetStringMBWithAllocator(from,
													 NULL, /* use malloc */
													 NULL, /* not use length */
													 &buf);
        if (OPENSOAP_SUCCEEDED(ret)) {
			*(long *)to_value = atol(buf);
			free(buf);
        }
    }
    
    return ret;
}

/*
=begin
--- function#OPENSOAP_SERIALIZER_NAME(short)(from_value, to)
    OpenSOAP Serializer of Short Type

    :Parameters
      :[in]  void * ((|from_value|))
        Serializer Object Assign Pointer
      :[out] OpenSOAPStringPtr ((|to|))
	    Result Value Pointer
    :Return Value
      :int
	    Return Code Is ErrorCode
    :Note
      16bit
=end
 */
OPENSOAP_SERIALIZER_DEFINITION(short) {
    int ret = OPENSOAP_PARAMETER_BADVALUE;

    if (from_value && to) {
        short val = *(short *)from_value;
        ret = OpenSOAPStringFormatWC(to,
                                     L"%d",
                                     val);
    }
    
    return ret;
}

/*
=begin
--- function#OPENSOAP_DESERIALIZER_NAME(short)(from, to_value)
    OpenSOAP Deserializer of Short Type

    :Parameters
      :[in]  OpenSOAPStringPtr ((|from|))
        Deserializer Object Assign Pointer
      :[out] void * ((|to_value|))
	    Result Value Pointer
    :Return Value
      :int
	    Return Code Is ErrorCode
    :Note
      16bit
=end
 */
OPENSOAP_DESERIALIZER_DEFINITION(short) {
    int ret = OPENSOAP_PARAMETER_BADVALUE;

    if (from && to_value) {
		char *buf = NULL;
		ret = OpenSOAPStringGetStringMBWithAllocator(from,
													 NULL, /* use malloc */
													 NULL, /* not use length */
													 &buf);
        if (OPENSOAP_SUCCEEDED(ret)) {
			*(short *)to_value = (short)atoi(buf);
			free(buf);
        }
    }
    return ret;
}

/*
=begin
--- function#OPENSOAP_SERIALIZER_NAME(string)(from_value, to)
    OpenSOAP Serializer of String Type

    :Parameters
      :[in]  void * ((|from_value|))
        Serializer Object Assign Pointer
      :[out] OpenSOAPStringPtr ((|to|))
	    Result Value Pointer
    :Return Value
      :int
	    Return Code Is ErrorCode
    :Note
      String
=end
 */
OPENSOAP_SERIALIZER_DEFINITION(string) {
    int ret = OPENSOAP_PARAMETER_BADVALUE;

    if (from_value && to) {
        OpenSOAPStringPtr from_str = *(OpenSOAPStringPtr *)from_value;
        ret = OpenSOAPStringDuplicate(from_str, &to);
    }
    return ret;
}

/*
=begin
--- function#OPENSOAP_DESERIALIZER_NAME(string)(from, to_value)
    OpenSOAP Deserializer of String Type

    :Parameters
      :[in]  OpenSOAPStringPtr ((|from|))
        Deserializer Object Assign Pointer
      :[out] void * ((|to_value|))
	    Result Value Pointer
    :Return Value
      :int
	    Return Code Is ErrorCode
    :Note
      String
=end
 */
OPENSOAP_DESERIALIZER_DEFINITION(string) {
    int ret = OPENSOAP_PARAMETER_BADVALUE;
	OpenSOAPStringPtr *toStr = (OpenSOAPStringPtr *)to_value;

    if (from && toStr
#if defined(STRING_DEDESIRALIZER_CHECKED)
		&& *toStr
#endif		
		) {
        ret = OpenSOAPStringDuplicate(from, toStr);
    }
    return ret;
}

/*-------------------------------*/
/* float serializer/deserializer */
/*-------------------------------*/

/*
=begin
--- function#OPENSOAP_SERIALIZER_NAME(float)(from_value, to)
    OpenSOAP Serializer of float Type

    :Parameters
      :[in]  void * ((|from_value|))
        Serializer Object Assign Pointer
      :[out] OpenSOAPStringPtr ((|to|))
	    Result Value Pointer
    :Return Value
      :int
	    Return Code Is ErrorCode
    :Note
      32bit floating point number.
=end
 */
OPENSOAP_SERIALIZER_DEFINITION(float) {
    int ret = OPENSOAP_PARAMETER_BADVALUE;

    if (from_value && to) {
        float val = *(float *)from_value;
        ret = OpenSOAPStringFormatWC(to,
                                     L"%f",
                                     val);
    }
    return ret;
}

/*
=begin
--- function#OPENSOAP_DESERIALIZER_NAME(float)(from, to_value)
    OpenSOAP Deserializer of float Type

    :Parameters
      :[in]  OpenSOAPStringPtr ((|from|))
        Deserializer Object Assign Pointer
      :[out] void * ((|to_value|))
	    Result Value Pointer
    :Return Value
      :int
	    Return Code Is ErrorCode
    :Note
      32bit floating point number.
=end
 */
OPENSOAP_DESERIALIZER_DEFINITION(float) {
    int ret = OPENSOAP_PARAMETER_BADVALUE;

    if (from && to_value) {
		char *buf = NULL;
		ret = OpenSOAPStringGetStringMBWithAllocator(from,
													 NULL, /* use malloc */
													 NULL, /* not use length */
													 &buf);
        if (OPENSOAP_SUCCEEDED(ret)) {
			*(float *)to_value = (float)atof(buf);
			free(buf);
        }
    }
    return ret;
}

/*------------------------------*/
/* byte serializer/deserializer */
/*------------------------------*/

/*
=begin
--- function#OPENSOAP_SERIALIZER_NAME(byte)(from_value, to)
    OpenSOAP Serializer of byte Type

    :Parameters
      :[in]  void * ((|from_value|))
        Serializer Object Assign Pointer. This type is required
		 signed char *.
      :[out] OpenSOAPStringPtr ((|to|))
	    Result Value Pointer
    :Return Value
      :int
	    Return Code Is ErrorCode
    :Note
      byte is integer type and it's region is [-128, 127].
=end
 */
OPENSOAP_SERIALIZER_DEFINITION(byte) {
    int ret = OPENSOAP_PARAMETER_BADVALUE;

    if (from_value && to) {
        int val = *(signed char *)from_value;
        ret = OpenSOAPStringFormatWC(to,
                                     L"%d",
                                     val);
    }
    
    return ret;
}

/*
=begin
--- function#OPENSOAP_DESERIALIZER_NAME(byte)(from, to_value)
    OpenSOAP Deserializer of byte Type

    :Parameters
      :[in]  OpenSOAPStringPtr ((|from|))
        Deserializer Object Assign Pointer
      :[out] void * ((|to_value|))
	    Result Value Pointer
    :Return Value
      :int
	    Return Code Is ErrorCode
    :Note
      byte is integer and it's region is [-128, 127].
=end
 */
OPENSOAP_DESERIALIZER_DEFINITION(byte) {
    int ret = OPENSOAP_PARAMETER_BADVALUE;

    if (from && to_value) {
		char *buf = NULL;
		ret = OpenSOAPStringGetStringMBWithAllocator(from,
													 NULL, /* use malloc */
													 NULL, /* not use length */
													 &buf);
        if (OPENSOAP_SUCCEEDED(ret)) {
			*(signed char *)to_value = (signed char)atoi(buf);
			free(buf);
        }
    }
    return ret;
}

/*----------------------------------*/
/* dateTime serializer/deserializer */
/*----------------------------------*/

#include <time.h>

/*
=begin
--- function#OPENSOAP_SERIALIZER_NAME(dateTime)(from_value, to)
    OpenSOAP Serializer of dateTime Type

    :Parameters
      :[in]  void * ((|from_value|))
        Serializer Object Assign Pointer. This type is required
		 signed char *.
      :[out] OpenSOAPStringPtr ((|to|))
	    Result Value Pointer
    :Return Value
      :int
	    Return Code Is ErrorCode
=end
 */
OPENSOAP_SERIALIZER_DEFINITION(dateTime) {
	static const char XMLdateTimeFormat[] = "%Y-%m-%dT%H:%M:%SZ";
    int ret = OPENSOAP_PARAMETER_BADVALUE;

    if (from_value && to) {
        const struct tm *val = (const struct tm *)from_value;
#define BUFDEFSIZE 22
#define BUFGRAWSIZE 4
        size_t  maxlen = BUFDEFSIZE;
        char *buf = NULL;
        char *old_buf = NULL;

        for (; buf = realloc(old_buf, (maxlen + 1));
             old_buf = buf) {
            size_t len = strftime(buf, maxlen, XMLdateTimeFormat, val);
            if (len < maxlen && len > 0) {
                break;
            }
            maxlen += BUFGRAWSIZE;
        }
        if (!buf) {
            free(old_buf);
            ret = OPENSOAP_MEM_BADALLOC;
        }
        else {
            ret = OpenSOAPStringSetStringMB(to, buf);
            free(buf);
        }
#undef BUFGRAWSIZE
#undef BUFDEFSIZE
    }
    
    return ret;
}

static
int
isTermChars(char chr,
			const char *termChars,
			size_t termCharsSize) {
	int ret = 0;

	if (termChars && termCharsSize) {
		const char *tc = termChars;
		const char * const tcEnd = termChars + termCharsSize;
		for (; tc != tcEnd; ++tc) {
			if (*tc == chr) {
				ret = 1;
				break;
			}
		}
	}

	return ret;
}

static
int
getLongValueWithTermChars(/* [in, out]  */ const char **str,
						  /* [in]  */ const char *strEnd,
						  /* [in]  */ const char *termChars,
						  /* [in]  */ size_t termCharsSize,
						  /* [out] */ long *value) {
	int ret = OPENSOAP_PARAMETER_BADVALUE;

	if (str && *str && *str < strEnd && termChars && termCharsSize && value) {
		const char *strI   = *str;
		ret = OPENSOAP_NO_ERROR;
		*value = 0;
		for (; strI != strEnd
				 && !isTermChars(*strI, termChars, termCharsSize); ++strI) {
			if (!isdigit(*strI)) {
				ret = OPENSOAP_PARAMETER_BADVALUE;
				break;
			}
			*value *= 10;
			*value += (*strI - '0');
		}
		*str = strI;
	}

	return ret;
}

static
int
OpenSOAPDeserializerMBStrToDate(/* [in, out] */ const char **str,
								/* [in]      */ const char *strEnd,
								/* [out]     */ struct tm *toValue) {
	static const char YEAR_MONTH_TERM_CHARS[] = "-";
	static size_t YEAR_MONTH_TERM_CHARS_SIZE
		= sizeof(YEAR_MONTH_TERM_CHARS);
	static const char DAY_TERM_CHARS[] = "T";
	static size_t DAY_TERM_CHARS_SIZE
		= sizeof(DAY_TERM_CHARS);
	int ret = OPENSOAP_PARAMETER_BADVALUE;

	if (str && *str && *str < strEnd && toValue) {
		long year = 0;
		ret = getLongValueWithTermChars(str,
										strEnd,
										YEAR_MONTH_TERM_CHARS,
										YEAR_MONTH_TERM_CHARS_SIZE,
										&year);
		if (OPENSOAP_SUCCEEDED(ret)) {
			long month = 0;
			++(*str);
			ret = getLongValueWithTermChars(str,
											strEnd,
											YEAR_MONTH_TERM_CHARS,
											YEAR_MONTH_TERM_CHARS_SIZE,
											&month);
			if (OPENSOAP_SUCCEEDED(ret)) {
				long day = 0;
				++(*str);
				ret = getLongValueWithTermChars(str,
												strEnd,
												DAY_TERM_CHARS,
												DAY_TERM_CHARS_SIZE,
												&day);
				if (OPENSOAP_SUCCEEDED(ret)) {
					if (month <= 0 || month > 12
						|| day <= 0 || day > 31) {
						ret = OPENSOAP_PARAMETER_BADVALUE;
					}
					else {
						toValue->tm_year = (year - 1900);
						toValue->tm_mon  = month - 1;
						toValue->tm_mday = day;
					}
				}
			}
		}
	}

	return ret;
}

static
int
OpenSOAPDeserializerMBStrToHourMin(/* [in, out] */ const char **str,
								   /* [in]      */ const char *strEnd,
								   /* [out]     */ struct tm *toValue) {
	static const char HOUR_MIN_TERM_CHARS[]	= ":";
	static size_t HOUR_MIN_TERM_CHARS_SIZE	= sizeof(HOUR_MIN_TERM_CHARS);
	int ret = OPENSOAP_PARAMETER_BADVALUE;

	if (str && *str && *str < strEnd && toValue) {
		long hour = 0;
		ret = getLongValueWithTermChars(str,
										strEnd,
										HOUR_MIN_TERM_CHARS,
										HOUR_MIN_TERM_CHARS_SIZE,
										&hour);
		if (OPENSOAP_SUCCEEDED(ret)) {
			long minutes = 0;
			++(*str);
			ret = getLongValueWithTermChars(str,
											strEnd,
											HOUR_MIN_TERM_CHARS,
											HOUR_MIN_TERM_CHARS_SIZE,
											&minutes);
			if (OPENSOAP_SUCCEEDED(ret)) {
				if (hour < 0 || hour > 23
					|| minutes < 0 || minutes > 59) {
					ret = OPENSOAP_PARAMETER_BADVALUE;
				}
				else {
					toValue->tm_hour = hour;
					toValue->tm_min  = minutes;
				}
			}
		}
	}

	return ret;
}

static const char SECOND_TERM_CHARS[]	= ".-Z+";
static const size_t SECOND_TERM_CHARS_SIZE	= sizeof(SECOND_TERM_CHARS);
static const char *PRE_TZ_TERM_CHARS	= SECOND_TERM_CHARS + 1;
static const size_t PRE_TZ_TERM_CHARS_SIZE	= sizeof(SECOND_TERM_CHARS) - 1;

static
int
OpenSOAPDeserializerMBStrToSec(/* [in, out] */ const char **str,
							   /* [in]      */ const char *strEnd,
							   /* [in. out] */ struct tm *toValue) {
	int ret = OPENSOAP_PARAMETER_BADVALUE;

	if (str && *str && *str < strEnd && toValue) {
		long sec = 0;
		ret = getLongValueWithTermChars(str,
										strEnd,
										SECOND_TERM_CHARS,
										SECOND_TERM_CHARS_SIZE,
										&sec);
		if (OPENSOAP_SUCCEEDED(ret)) {
			if (sec <= 0 || sec > 61) {
				ret = OPENSOAP_PARAMETER_BADVALUE;
			}
			else {
				toValue->tm_sec  = sec;
				if (**str == SECOND_TERM_CHARS[0]) { /* is point */
					long dummy = 0;
					++(*str);
					ret = getLongValueWithTermChars(str,
													strEnd,
													PRE_TZ_TERM_CHARS,
													PRE_TZ_TERM_CHARS_SIZE,
													&dummy);
				}
			}
		}
	}

	return ret;
}

static
int
OpenSOAPDeserializerMBStrToModifyTZ(/* [in, out] */ const char **str,
									/* [in]      */ const char *strEnd,
									/* [in. out] */ struct tm *toValue) {
	int ret = OPENSOAP_PARAMETER_BADVALUE;

	if (str && *str && *str < strEnd && toValue) {
		const char *strI = *str;
		const char *preTzChar = strchr(PRE_TZ_TERM_CHARS, **str);
		int sign = preTzChar ? -(preTzChar - PRE_TZ_TERM_CHARS - 1) : 0;
		if (sign) {
			struct tm tz;
			++(*str);
			ret = OpenSOAPDeserializerMBStrToHourMin(str, strEnd, &tz);
			if (OPENSOAP_SUCCEEDED(ret)) {
				time_t mktimeRet = (time_t)(-1);
				toValue->tm_hour += sign * tz.tm_hour;
				toValue->tm_min  += sign * tz.tm_min;
				mktimeRet = mktime(toValue);
				if (mktimeRet == (time_t)(-1)) {
					ret = OPENSOAP_PARAMETER_BADVALUE;
				}
			}
		}
		else {
			ret = OPENSOAP_NO_ERROR;
		}
	}

	return ret;
}

/*
=begin
--- function#OPENSOAP_DESERIALIZER_NAME(dateTime)(from, to_value)
    OpenSOAP Deserializer of dateTime Type

    :Parameters
      :[in]  OpenSOAPStringPtr ((|from|))
        Deserializer Object Assign Pointer
      :[out] void * ((|to_value|))
	    Result Value Pointer
    :Return Value
      :int
	    Return Code Is ErrorCode
=end
 */
OPENSOAP_DESERIALIZER_DEFINITION(dateTime) {
	int ret = OPENSOAP_PARAMETER_BADVALUE;

	if (from && to_value) {
		char *buf = NULL;
		size_t len = 0;
		ret = OpenSOAPStringGetStringMBWithAllocator(from,
													 NULL, /* use malloc */
													 &len,
													 &buf);
		if (OPENSOAP_SUCCEEDED(ret)) {
			const char * const bufEnd = buf + len + 1;
			const char *bufI = buf;
			struct tm *value = to_value;

			ret = OpenSOAPDeserializerMBStrToDate(&bufI,
												  bufEnd,
												  value);
			if (OPENSOAP_SUCCEEDED(ret)) {
				++bufI;
				ret = OpenSOAPDeserializerMBStrToHourMin(&bufI,
														 bufEnd,
														 value);
				if (OPENSOAP_SUCCEEDED(ret)) {
					++bufI;
					ret = OpenSOAPDeserializerMBStrToSec(&bufI,
														 bufEnd,
														 value);
					if (OPENSOAP_SUCCEEDED(ret)) {
						ret = OpenSOAPDeserializerMBStrToModifyTZ(&bufI,
																  bufEnd,
																  value);
					}
				}
			}
			free(buf);
		}
	}
	
	return ret;
}

/*
=begin
= Serializer registry code.
=end
 */

#include <OpenSOAP/StringHash.h>

typedef struct {
    OpenSOAPSerializerFunc      serializer;
    OpenSOAPDeserializerFunc    deserializer;
} StaticOpenSOAPSerializerPair, *StaticOpenSOAPSerializerPairPtr;

static
OpenSOAPStringHashPtr   serializerTable = NULL;

/*
=begin
--- function#createSerializerPair(serializer, deserializer)
    OpenSOAP create serializer pair

    :Parameters
      :OpenSOAPSerializerFunc [in] ((|serializer|))
        Serializer Function Program Pointer
      :OpenSOAPDeserializerFunc [out] ((|deserializer|))
        Deserializer Function Program Pointer
    :Return Value
      :StaticOpenSOAPSerializerPairPtr
	:Note
	  This Function Is Static Function
=end
 */
static
StaticOpenSOAPSerializerPairPtr
createSerializerPair(OpenSOAPSerializerFunc /* [in] */ serializer,
					 OpenSOAPDeserializerFunc /* [in] */ deserializer) {
	StaticOpenSOAPSerializerPairPtr ret
		= malloc(sizeof(StaticOpenSOAPSerializerPair));
	if (ret) {
		ret->serializer = serializer;
		ret->deserializer = deserializer;
	}
   
	return ret;
}

/*
=begin
--- function#OpenSOAPSerializerRegisterString(soap_typename, serializer, deserializer)
    Serializer and deserializer register function.

    :Parameters
      :[in] OpenSOAPStringPtr ((|soap_typename|))
        typename.
      :[in] OpenSOAPSerializerFunc ((|serializer|))
        serializer
      :[in] OpenSOAPDeserializerFunc ((|deserializer|))
        deserializer

    :Return Value
      :int
	    Return Code Is ErrorCode
	:Note
	  This Function Is Static Function
=end
 */
extern
int
OPENSOAP_API
OpenSOAPSerializerRegisterString(/* [in] */ OpenSOAPStringPtr soap_typename,
								 /* [in] */ OpenSOAPSerializerFunc serializer,
								 /* [in] */ OpenSOAPDeserializerFunc deserializer) {
    int ret = OPENSOAP_PARAMETER_BADVALUE;

    if (soap_typename && serializer && deserializer) {
        if (!serializerTable) {
            ret = OpenSOAPStringHashCreate(&serializerTable);
            if (OPENSOAP_FAILED(ret)) {
                serializerTable = NULL;
            }
        }
        if (serializerTable) {
            StaticOpenSOAPSerializerPairPtr serializer_pair
                = createSerializerPair(serializer,
                                       deserializer);
            if (!serializer_pair) {
                ret = OPENSOAP_MEM_BADALLOC;
            }
            else {
                ret = OpenSOAPStringHashSetValue(serializerTable,
                                                 soap_typename,
                                                 serializer_pair);
            }
        }
    }

    return ret;
}

/*
=begin
--- function#OpenSOAPSerializerRegisterMB(soap_typename, serializer, deserializer)
    Serializer and deserializer register function.(MB)

    :Parameters
      :[in] const char *((|soap_typename|))
        typename.
      :[in] OpenSOAPSerializerFunc ((|serializer|))
        serializer
      :[in] OpenSOAPDeserializerFunc ((|deserializer|))
        deserializer
    :Return Value
      :int
	    Return Code Is ErrorCode
=end
 */
extern
int
OPENSOAP_API
OpenSOAPSerializerRegisterMB(/* [in] */ const char *soap_typename,
                             /* [in] */ OpenSOAPSerializerFunc serializer,
                             /* [in] */ OpenSOAPDeserializerFunc deserializer) {
    int ret = OPENSOAP_PARAMETER_BADVALUE;

    if (soap_typename && *soap_typename) {
        OpenSOAPStringPtr   tname = NULL;
        ret = OpenSOAPStringCreateWithMB(soap_typename, &tname);
        if (OPENSOAP_SUCCEEDED(ret)) {
            ret = OpenSOAPSerializerRegisterString(tname,
												   serializer,
												   deserializer);
            OpenSOAPStringRelease(tname);
        }
    }

    return ret;
}

/*
=begin
--- function#OpenSOAPSerializerRegisterWC(soap_typename, serializer, deserializer)
    Serializer and deserializer register function.(WC)

    :Parameters
      : [in] const wchar_t *((|soap_typename|))
        typename.
      :[in] OpenSOAPSerializerFunc ((|serializer|))
        serializer
      :[in] OpenSOAPDeserializerFunc ((|deserializer|))
        deserializer
    :Return Value
      :int
	    Return Code Is ErrorCode
=end
 */
extern
int
OPENSOAP_API
OpenSOAPSerializerRegisterWC(/* [in] */ const wchar_t *soap_typename,
                             /* [in] */ OpenSOAPSerializerFunc serializer,
                             /* [in] */ OpenSOAPDeserializerFunc deserializer) {
    int ret = OPENSOAP_PARAMETER_BADVALUE;

    if (soap_typename && *soap_typename) {
        OpenSOAPStringPtr   tname = NULL;
        ret = OpenSOAPStringCreateWithWC(soap_typename, &tname);
        if (OPENSOAP_SUCCEEDED(ret)) {
            ret = OpenSOAPSerializerRegisterString(tname,
												   serializer,
												   deserializer);
            OpenSOAPStringRelease(tname);
        }
    }

    return ret;
}

/*
=begin
--- function#OpenSOAPGetSerializerPair(soap_typename, serializer_pair)
    Get registered serializer.

    :Parameters
      :[in]  OpenSOAPStringPtr ((|soap_typename|))
        typename.
      :[out] StaticOpenSOAPSerializerPairPtr * ((|serializer_pair|))
        serializer
    :Return Value
      :int
	    Return Code Is ErrorCode
	:Note
	  This Function Is Static Function
=end
 */
static
/* extern */
int
/* OPENSOAP_API */
OpenSOAPGetSerializerPair(/* [in] */ OpenSOAPStringPtr soap_typename,
                          /* [out] */ StaticOpenSOAPSerializerPairPtr *serializer_pair) {
    int ret = OPENSOAP_PARAMETER_BADVALUE;

    if (soap_typename && serializer_pair) {
        ret = OpenSOAPStringHashGetValue(serializerTable,
                                         soap_typename,
                                         (void **)serializer_pair);
    }
    
    return ret;
}

/*
=begin
--- function#OpenSOAPGetSerializer(soap_typename, serializer)
    Get registered serializer.(MB)

    :Parameters
      :[in]  OpenSOAPStringPtr ((|soap_typename|))
        typename.
      :[out] OpenSOAPSerializerFunc * ((|serializer|))
        serializer
    :Return Value
      :int
	    Return Code Is ErrorCode
=end
 */
extern
int
OPENSOAP_API
OpenSOAPGetSerializer(/* [in] */  OpenSOAPStringPtr soap_typename,
                      /* [out] */ OpenSOAPSerializerFunc *serializer) {
    int ret = OPENSOAP_PARAMETER_BADVALUE;
    if (soap_typename && serializer) {
        StaticOpenSOAPSerializerPairPtr serializer_pair = NULL;
        ret = OpenSOAPGetSerializerPair(soap_typename,
                                        &serializer_pair);
        if (OPENSOAP_SUCCEEDED(ret)) {
            *serializer = serializer_pair->serializer;
        }
    }
    
    return ret;
}

/*
=begin
--- function#OpenSOAPGetSerializerMB(soap_typename, serializer)
    Get registered serializer.(MB)

    :Parameters
      :[in]  const char * ((|soap_typename|))
        typename.
      :[out] OpenSOAPSerializerFunc * ((|serializer|))
        serializer
    :Return Value
      :int
	    Return Code Is ErrorCode
=end
 */
extern
int
OPENSOAP_API
OpenSOAPGetSerializerMB(/* [in] */ const char *soap_typename,
                        /* [out] */ OpenSOAPSerializerFunc *serializer) {
    OpenSOAPStringPtr   tname = NULL;
    int ret = OpenSOAPStringCreateWithMB(soap_typename, &tname);
    if (OPENSOAP_SUCCEEDED(ret)) {
        ret = OpenSOAPGetSerializer(tname, serializer);

        OpenSOAPStringRelease(tname);
    }

    return ret;
}

/*
=begin
--- function#OpenSOAPGetSerializerWC(soap_typename, serializer)
    Get registered serializer.(WC)

    :Parameters
      :[in]  const wchar_t * ((|soap_typename|))
        typename.
      :[out] OpenSOAPSerializerFunc * ((|serializer|))
        serializer
    :Return Value
      :int
	    Return Code Is ErrorCode
=end
 */
extern
int
OPENSOAP_API
OpenSOAPGetSerializerWC(/* [in] */ const wchar_t *soap_typename,
                        /* [out] */ OpenSOAPSerializerFunc *serializer) {
    OpenSOAPStringPtr   tname = NULL;
    int ret = OpenSOAPStringCreateWithWC(soap_typename, &tname);
    if (OPENSOAP_SUCCEEDED(ret)) {
        ret = OpenSOAPGetSerializer(tname, serializer);

        OpenSOAPStringRelease(tname);
    }

    return ret;
}

/*
=begin
--- function#OpenSOAPGetDeserializer(soap_typename, deserializer)
    Get registered serializer.(OpenSOAPString)

    :Parameters
      :[in]  OpenSOAPStringPtr ((|soap_typename|))
        typename.
      :[out] OpenSOAPDeserializerFunc * ((|deserializer|))
        deserializer
    :Return Value
      :int
	    Return Code Is ErrorCode
=end
 */
extern
int
OPENSOAP_API
OpenSOAPGetDeserializer(/* [in] */  OpenSOAPStringPtr soap_typename,
                        /* [out] */ OpenSOAPDeserializerFunc *deserializer) {
    int ret = OPENSOAP_PARAMETER_BADVALUE;
    if (soap_typename && deserializer) {
        StaticOpenSOAPSerializerPairPtr serializer_pair = NULL;
        ret = OpenSOAPGetSerializerPair(soap_typename,
                                        &serializer_pair);
        if (OPENSOAP_SUCCEEDED(ret)) {
            *deserializer = serializer_pair->deserializer;
        }
    }
    
    return ret;
}

/*
=begin
--- function#OpenSOAPGetDeserializerMB(soap_typename, deserializer)
    Get registered serializer.(MB)

    :Parameters
      :[in]  const char * ((|soap_typename|))
        typename.
      :[out] OpenSOAPDeserializerFunc * ((|deserializer|))
        deserializer
    :Return Value
      :int
	    Return Code Is ErrorCode
=end
 */
extern
int
OPENSOAP_API
OpenSOAPGetDeserializerMB(/* [in] */ const char *soap_typename,
                          /* [out] */ OpenSOAPDeserializerFunc *deserializer) {
    OpenSOAPStringPtr   tname = NULL;
    int ret = OpenSOAPStringCreateWithMB(soap_typename, &tname);
    if (OPENSOAP_SUCCEEDED(ret)) {
        ret = OpenSOAPGetDeserializer(tname, deserializer);

        OpenSOAPStringRelease(tname);
    }

    return ret;
}

/*
=begin
--- function#OpenSOAPGetSerializerWC(soap_typename, deserializer)
    Get registered serializer.(WC)

    :Parameters
      :[in]  const wchar_t * ((|soap_typename|))
        typename.
      :[out] OpenSOAPDeserializerFunc * ((|deserializer|))
        deserializer
    :Return Value
      :int
	    Return Code Is ErrorCode
=end
 */
extern
int
OPENSOAP_API
OpenSOAPGetDeserializerWC(/* [in] */  const wchar_t *soap_typename,
                          /* [out] */ OpenSOAPDeserializerFunc *deserializer) {
    OpenSOAPStringPtr   tname = NULL;
    int ret = OpenSOAPStringCreateWithWC(soap_typename, &tname);
    if (OPENSOAP_SUCCEEDED(ret)) {
        ret = OpenSOAPGetDeserializer(tname, deserializer);

        OpenSOAPStringRelease(tname);
    }

    return ret;
}

typedef struct {
    const wchar_t *soapTypename;
    OpenSOAPSerializerFunc serializer;
    OpenSOAPDeserializerFunc deserializer;
} OpenSOAPSerializerRegistInfo;

typedef OpenSOAPSerializerRegistInfo *OpenSOAPSerializerRegistInfoPtr;

static
OpenSOAPSerializerRegistInfo
SerializerRegisterInfos[] = {
	OPENSOAP_SERIALIZER_REGIST_INFO_ITEM(boolean),
	OPENSOAP_SERIALIZER_REGIST_INFO_ITEM(base64Binary),
	OPENSOAP_SERIALIZER_REGIST_INFO_ITEM(double),
	OPENSOAP_SERIALIZER_REGIST_INFO_ITEM(int),
	OPENSOAP_SERIALIZER_REGIST_INFO_ITEM(short),
	OPENSOAP_SERIALIZER_REGIST_INFO_ITEM(string),
	OPENSOAP_SERIALIZER_REGIST_INFO_ITEM(float),
	OPENSOAP_SERIALIZER_REGIST_INFO_ITEM(byte),
	OPENSOAP_SERIALIZER_REGIST_INFO_ITEM(dateTime),
    { NULL, NULL, NULL}
};
    
/*
=begin
--- function#OpenSOAPSerializerRegistDefaults()

    :Parameters
    :Return Value
      :int
	    Return Code Is ErrorCode
=end
 */
extern
int
OPENSOAP_API
OpenSOAPSerializerRegistDefaults(void) {
    int ret = OPENSOAP_NO_ERROR;
    OpenSOAPSerializerRegistInfoPtr i
        = SerializerRegisterInfos;
    for (;OPENSOAP_SUCCEEDED(ret) && i->soapTypename; ++i) {
        ret = OpenSOAPSerializerRegisterWC(i->soapTypename,
                                           i->serializer,
                                           i->deserializer);
    }

    return ret;
}
